/* 2^n modular arithmetic.

   Copyright 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"
#include "mpm.h"
#include "mpm-impl.h"


/* A residue is msize many limbs, but has 2*msize allocated so as to avoid
   some temporaries.  Extra bits above MBITS(m) are not masked, so the
   representation is not unique.  */


/* change this to "#define TRACE(x) x" for diagnostics */
#define TRACE(x)


#define MMASK(m)  MNORM_RSHIFT_MASK(m)

#define ASSERT_MPR_2EXP(r,m)            \
  do {                                  \
    ASSERT_MPN (RPTR(r), MRSIZE(m));    \
  } while (0)

/* least significant 1 bit, must have n!=0 */
#define LOW_ONE_BIT(n)   ((n)&-(n))


static void
mpr_2exp__add (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  ASSERT_MPR_2EXP (x, m);
  ASSERT_MPR_2EXP (y, m);

  mpn_add_n (RPTR(w), RPTR(x), RPTR(y), MRSIZE(m));

  ASSERT_MPR_2EXP (y, m);
}

static void
mpr_2exp__add_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_size_t  msize;
  mp_limb_t  c;

  ASSERT_MPR_2EXP (x, m);

  wp = RPTR(w);
  xp = RPTR(x);
  msize = MRSIZE(m);

#if BITS_PER_ULONG > GMP_NUMB_BITS
  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      if (msize == 1)
        wp[0] = (xp[0] + y) & GMP_NUMB_MASK;
      else
        {
          mp_limb_t yy = y;
          MPN_ADD_2 (c, wp, xp, msize,
                     yy >> GMP_NUMB_BITS,
                     yy & GMP_NUMB_MASK);
        }
    }
  else
#endif
    {
      MPN_ADD_1 (c, wp, xp, msize, (mp_limb_t) y);
    }

  ASSERT_MPR_2EXP (w, m);
}

static void
mpr_2exp__addmul_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_size_t  msize;

  ASSERT_MPR_2EXP (w, m);
  ASSERT_MPR_2EXP (x, m);
  ASSERT (MRALLOC(m) >= MRSIZE(m)+1);

  wp = RPTR(w);
  xp = RPTR(x);
  msize = MRSIZE(m);

#if BITS_PER_ULONG > GMP_NUMB_BITS
  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mp_limb_t  yp[2], c;
      mp_limb_t  yy = y;
      ASSERT_CODE (wp[msize] = 0); /* so nail==0 */
      yp[0] = yy & GMP_NUMB_MASK;
      yp[1] = yy >> GMP_NUMB_BITS;
      MPN_ADDMUL_2 (c, wp, xp, msize, yp);
    }
  else
#endif
    {
      mpn_addmul_1 (wp, xp, msize, (mp_limb_t) y);
    }

  ASSERT_MPR_2EXP (w, m);
}

static void
mpr_2exp__div (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_ptr     sp, ip, qp, wp;
  mp_size_t  i, msize, dsize, csize;
  mp_srcptr  xp, yp, dp;
  int        shift;
  TMP_DECL (marker);

  ASSERT_MPR_2EXP (x, m);
  ASSERT_MPR_2EXP (y, m);

  TMP_MARK (marker);
  msize = MRSIZE(m);
  TMP_ALLOC_LIMBS_3 (sp,msize, ip,msize, qp,2*msize);

  xp = RPTR(x);
  yp = RPTR(y);
  if (UNLIKELY ((yp[0] & 1) == 0))
    DIVIDE_BY_ZERO;

  i = 0;
  for (;;)
    {
      if (yp[i] == 0)
        {
          if (xp[i] != 0)
            {
            div0:
              DIVIDE_BY_ZERO;
            }
        }
      i++;
      if (i >= msize)
        goto div0;  /* 0/0 */
    }

  count_trailing_zeros (shift, yp[i]);
  dsize = msize-i;
  if (shift != 0)
    {
      if (mpn_rshift (sp, yp+i, dsize, shift) != 0)
        goto div0;
      dp = sp;
      dsize -= (dp[dsize-1] == 0);
    }
  else
    dp = yp+i;

  mpn_modlimbs_invert (ip, dp, dsize);

  if (dsize == msize)
    {
      mpn_mullowhalf_n (qp, RPTR(x), ip, dsize);
      csize = msize;
    }
  else
    {
      mpn_mul_n (qp, RPTR(x), ip, dsize);
      csize = MIN (2*dsize, msize);
    }
  wp = RPTR(w);
  MPN_COPY (wp, qp, csize);
  MPN_ZERO (wp+csize, msize-csize);

  TMP_FREE (marker);
}

/* 1/2 not well defined for m even */
static void
mpr_2exp__div_2exp (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  ASSERT_MPR_2EXP (x, m);
  DIVIDE_BY_ZERO;
}

static int
mpr_2exp__equal (mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_size_t  i, msize;
  mp_srcptr  xp, yp;

  ASSERT_MPR_2EXP (x, m);

  msize = MRSIZE(m);
  xp = RPTR(x);
  yp = RPTR(y);

  for (i = 0; i < msize-1; i++)
    if (xp[i] != yp[i])
      return 0;

  return (((xp[msize-1] - yp[msize-1]) & MMASK(m)) == 0);
}

static int
mpr_2exp__equal_one (mpr_srcptr x, mpm_srcptr m)
{
  mp_size_t  i, msize;
  mp_srcptr  xp;
  mp_limb_t  xlow;

  ASSERT_MPR_2EXP (x, m);

  xp = RPTR(x);
  msize = MRSIZE(m);

  xlow = xp[0];
  if (msize == 1)
    return ((xlow & MMASK(m)) == 1);

  if (xlow != 1)
    return 0;

  for (i = 1; i < msize-1; i++)
    if (xp[i] != 0)
      return 0;

  return ((xp[msize-1] & MMASK(m)) == 0);
}

static int
mpr_2exp__equal_negone (mpr_srcptr x, mpm_srcptr m)
{
  mp_size_t  i, msize;
  mp_srcptr  xp;

  ASSERT_MPR_2EXP (x, m);

  msize = MRSIZE(m);
  xp = RPTR(x);

  for (i = 0; i < msize-1; i++)
    if (xp[i] != GMP_NUMB_MASK)
      return 0;

  return (((xp[msize-1] + 1) & MMASK(m)) == 0);
}

static int
mpr_2exp__equal_ui__1 (mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  ASSERT_MPR_2EXP (x, m);
  ASSERT (MRSIZE(m) == 1);

  return (((RPTR(x)[0] - y) & MMASK(m)) == 0);
}

static int
mpr_2exp__equal_ui__2up (mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  i, msize;
  mp_srcptr  xp;

  ASSERT_MPR_2EXP (x, m);
  ASSERT (MRSIZE(m) >= 2);

  xp = RPTR(x);

  if (xp[0] != ((mp_limb_t) y & GMP_NUMB_MASK))
    return 0;
  i = 1;

  msize = MRSIZE(m);

#if BITS_PER_ULONG > GMP_NUMB_BITS
  {
    mp_limb_t  xlimb;
    xlimb = xp[1];
    y >>= GMP_NUMB_BITS;
    if (UNLIKELY (msize == 2))
      return (((xlimb - y) & MMASK(m)) == 0);

    if (xlimb != y)
      return 0;
  }
  i = 2;
#endif

  for ( ; i < msize-1; i++)
    if (xp[i] != 0)
      return 0;

  if ((xp[msize-1] & MMASK(m)) != 0)
    return 0;

  return 1;
}

static int
mpr_2exp__equal_zero (mpr_srcptr x, mpm_srcptr m)
{
  mp_size_t  i, msize;
  mp_srcptr  xp;

  ASSERT_MPR_2EXP (x, m);

  xp = RPTR(x);
  msize = MRSIZE(m);

  for (i = 0; i < msize-1; i++)
    if (xp[i] != 0)
      return 0;

  if ((xp[msize-1] & MMASK(m)) != 0)
    return 0;

  return 1;
}

static void
mpr_2exp__gcd (mpz_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_srcptr  xp;
  mp_ptr     wp;
  mp_size_t  msize, i;
  mp_limb_t  xlimb;

  ASSERT_MPR_2EXP (x, m);

  msize = MRSIZE(m);
  xp = RPTR(x);
  i = 0;
  for (;;)
    {
      xlimb = xp[i];
      if (i == msize-1)
        {
          xlimb |= MMASK(m) ^ GMP_NUMB_MASK;
          if (xlimb == 0)
            {
              i++;
              xlimb = 1;
            }
          break;
        }
      if (xlimb != 0)
        break;
      i++;
    }

  SIZ(w) = i+1;
  MPZ_REALLOC_PTR (wp, w, i+1);

  ASSERT (xlimb != 0);
  wp[i] = LOW_ONE_BIT (xlimb);

  MPN_ZERO (wp, i);

  ASSERT_MPR_GCD (w, m);
}

static unsigned long
mpr_2exp__get_ui__1 (mpr_srcptr x, mpm_srcptr m)
{
  ASSERT_MPR_2EXP (x, m);
  return RPTR(x)[0] & MMASK(m);
}

#if BITS_PER_ULONG <= GMP_NUMB_BITS

#define mpr_2exp__get_ui                \
  (msize == 1 ? mpr_2exp__get_ui__1     \
   :            mpr_2exp__get_ui__2up)

static unsigned long
mpr_2exp__get_ui__2up (mpr_srcptr x, mpm_srcptr m)
{
  ASSERT_MPR_2EXP (x, m);
  return RPTR(x)[0];
}

#else

#define mpr_2exp__get_ui                        \
  (msize == 1   ? mpr_2exp__get_ui__1           \
   : msize == 2 ? mpr_2exp__get_ui__2           \
   :              mpr_2exp__get_ui__3up)

static unsigned long
mpr_2exp__get_ui__2 (mpr_srcptr x, mpm_srcptr m)
{
  mp_srcptr  xp;
  ASSERT_MPR_2EXP (x, m);
  xp = RPTR(x);
  return xp[0] + ((xp[1] & MMASK(m)) << GMP_NUMB_BITS);
}

static unsigned long
mpr_2exp__get_ui__3up (mpr_srcptr x, mpm_srcptr m)
{
  mp_srcptr  xp;
  ASSERT_MPR_2EXP (x, m);
  xp = RPTR(x);
  return xp[0] + (xp[1] << GMP_NUMB_BITS);
}
#endif

static void
mpr_2exp__get_z (mpz_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_size_t  msize, i;
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_limb_t  limb;

  ASSERT_MPR_2EXP (x, m);

  xp = RPTR(x);
  msize = MRSIZE(m);

  /* find the highest non-zero limb, if any */
  limb = xp[msize-1] & MMASK(m);
  i = msize-1;
  for (;;)
    {
      if (LIKELY (limb != 0))
        break;
      if (UNLIKELY (i == 0))
        {
          SIZ(w) = 0;
          return;
        }
      i--;
      limb = xp[i];
    }

  MPZ_REALLOC_PTR (wp, w, i+1);
  SIZ(w) = i+1;
  wp[i] = limb;
  MPN_COPY (wp, xp, i);

  ASSERT_MPR_GET_Z (w, m);
}

static int
mpr_2exp__invert (mpr_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_ptr     wp, tp;
  mp_srcptr  xp;
  mp_size_t  msize;
  TMP_DECL (marker);

  /* inverse only exists if x odd */
  xp = RPTR(x);
  if (UNLIKELY ((xp[0] & 1) == 0))
    return 0;

  TMP_MARK (marker);
  wp = RPTR(w);
  msize = MRSIZE(m);
  if (w == x)
    {
      tp = TMP_ALLOC_LIMBS (msize);
      MPN_COPY (tp, xp, msize);
      xp = tp;
    }
  mpn_modlimbs_invert (wp, xp, msize);

  TMP_FREE (marker);
  ASSERT_MPR_2EXP (w, m);
  return 1;
}

static int
mpr_2exp__jacobi (mpr_srcptr x, mpm_srcptr m)
{
  /* not well-defined for m even */
  DIVIDE_BY_ZERO;
}

void
mpr_2exp__mul (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_ptr     wp;
  mp_srcptr  xp, yp;
  mp_size_t  msize;

  ASSERT_MPR_2EXP (x, m);
  ASSERT_MPR_2EXP (y, m);
  ASSERT (MRALLOC(m) >= 2*MRSIZE(m));

  if (w == x)
    {
      mp_ptr     tp;
      TMP_DECL (marker);
    overlap:
      TMP_MARK (marker);
      msize = MRSIZE(m);
      tp = TMP_ALLOC_LIMBS (msize);
      MPN_COPY (tp, RPTR(x), msize);

      wp = RPTR(w);
      yp = RPTR(y);
      if (x == y)
        mpn_sqrlowhalf_n (wp, tp, msize);
      else
        mpn_mullowhalf_n (wp, tp, yp, msize);
      TMP_FREE (marker);
    }
  else if (w == y)
    {
      MPR_SRCPTR_SWAP (x, y);
      goto overlap;
    }
  else
    {
      wp = RPTR(w);
      xp = RPTR(x);
      yp = RPTR(y);
      msize = MRSIZE(m);
      if (x == y)
        mpn_sqrlowhalf_n (wp, xp, msize);
      else
        mpn_mullowhalf_n (wp, xp, yp, msize);
    }

  ASSERT_MPR_2EXP (w, m);
}

static void
mpr_2exp__mul_2exp (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  ylimbs, msize;
  int        yshift;
  mp_ptr     wp;
  mp_srcptr  xp;

  ASSERT_MPR_2EXP (x, m);

  ylimbs = y / GMP_NUMB_BITS;
  msize = MRSIZE(m);
  wp = RPTR(w);

  ylimbs = MIN (ylimbs, msize);
  if (LIKELY (ylimbs < msize))
    {
      yshift = y % GMP_NUMB_BITS;
      xp = RPTR(x);
      if (yshift != 0)
        mpn_lshift (wp+ylimbs, xp, msize-ylimbs, yshift);
      else
        MPN_COPY_DECR (wp+ylimbs, xp, msize-ylimbs);
    }

  MPN_ZERO (wp, ylimbs);
}

static void
mpr_2exp__mul_si (mpr_ptr w, mpr_srcptr x, long y, mpm_srcptr m)
{
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_size_t  msize;
  mp_limb_t  c, ylimb;

  ASSERT_MPR_2EXP (x, m);

  xp = RPTR(x);
  wp = RPTR(w);
  msize = MRSIZE(m);

  ylimb = (y >= 0 ? y : - (mp_limb_t) y);

#if BITS_PER_ULONG > GMP_NUMB_BITS
  if (UNLIKELY (ylimb > GMP_NUMB_MAX))
    {
      mp_limb_t  yp[2];
      yp[0] = ylimb & GMP_NUMB_MASK;
      yp[1] = ylimb >> GMP_NUMB_BITS;
      MPN_MUL_2 (c, wp, xp, msize, yp);
    }
  else
#endif
    {
      mpn_mul_1 (wp, xp, msize, ylimb);
    }

  if (y < 0)
    {
      mpn_com_n (wp, wp, msize);
      MPN_ADD_1 (c, wp, wp, msize, CNST_LIMB(1));
    }

  ASSERT_MPR_2EXP (w, m);
}

static void
mpr_2exp__mul_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_size_t  msize;

  ASSERT_MPR_2EXP (x, m);
  ASSERT (MRALLOC(m) >= MRSIZE(m)+1);

  wp = RPTR(w);
  xp = RPTR(x);
  msize = MRSIZE(m);

#if BITS_PER_ULONG > GMP_NUMB_BITS
  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mp_limb_t  yp[2], c;
      yp[0] = (mp_limb_t) y & GMP_NUMB_MASK;
      yp[1] = (mp_limb_t) y >> GMP_NUMB_BITS;
      MPN_MUL_2 (c, wp, xp, msize, yp);
    }
  else
#endif
    {
      mpn_mul_1 (wp, xp, msize, (mp_limb_t) y);
    }

  ASSERT_MPR_2EXP (w, m);
}

static void
mpr_2exp__neg (mpr_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;

  ASSERT_MPR_2EXP (x, m);

  wp = RPTR(w);
  msize = MRSIZE(m);
  mpn_com_n (wp, RPTR(x), msize);
  MPN_ADD_ONE_INPLACE_NOTEMPTY_IGNCARRY (wp, msize);
}

/* ENHANCE-ME: Could examine e to predict where the result will end up, in
   rp or tp, and avoid the final copy.

   ENHANCE-ME: Sliding window, perhaps with on-demand calculation of
   intermediate powers.  */

static void
mpr_2exp__pow (mpr_ptr r, mpr_srcptr b, mpz_srcptr e, mpm_srcptr m)
{
  mp_size_t  msize, tsize, esize, i;
  mp_ptr     rp, tp;
  mp_srcptr  bp, ep;
  mp_limb_t  elimb, mask;
  int        ezeros;
  TMP_DECL (marker);

  ASSERT_MPR_2EXP (b, m);

  TMP_MARK (marker);
  msize = MRSIZE(m);
  rp = RPTR(r);

  esize = SIZ(e);
  if (UNLIKELY (esize == 0))
    {
      /* b^0 == 1 */
      rp[0] = 1;
      MPN_ZERO (rp+1, msize-1);
      return;

    }

  bp = RPTR(b);

  tsize = 2*msize;
  if (esize < 0 || rp == bp)
    tsize += msize;

  tp = TMP_ALLOC_LIMBS (tsize);

  if (UNLIKELY (esize < 0))
    {
      if ((bp[0] & 1) == 0)
        DIVIDE_BY_ZERO;
      mpn_modlimbs_invert (tp, bp, msize);
      bp = tp;
      tsize -= msize;
      tp += msize;
      esize = -esize;
      goto copy;
    }
  else
    {
      if (rp == bp)
        {
          /* preserve original b */
          MPN_COPY (tp, bp, msize);
          bp = tp;
          tsize -= msize;
          tp += msize;
        }
      else
        {
        copy:
          MPN_COPY (rp, bp, msize);  /* start b^1 */
        }
    }
  ASSERT (tsize == 2*msize);

  ep = PTR(e);
  i = esize-1;
  elimb = ep[i];
  count_leading_zeros (ezeros, elimb);
  mask = CNST_LIMB(1) << (BITS_PER_MP_LIMB - 1 - ezeros);

  for (;;)
    {
      mask >>= 1;
      if (mask == 0)
        {
          i--;
          if (i < 0)
            break;
          elimb = ep[i];
          mask = GMP_NUMB_HIGHBIT;
        }
      TRACE (printf (" at i=%ld mask=%lX bit=%d\n",
                     i, mask, (elimb & mask) != 0));

      mpn_sqrlowhalf_n (tp, rp, msize);
      MP_PTR_SWAP (tp, rp);
      TRACE (mpn_trace (" double r", rp, msize));

      if ((elimb & mask) != 0)
        {
          mpn_mullowhalf_n (tp, rp, bp, msize);
          MP_PTR_SWAP (tp, rp);
          TRACE (mpn_trace (" step   r", rp, msize));
        }
    }

  if (rp != RPTR(r))
    MPN_COPY (RPTR(r), rp, msize);
  TMP_FREE (marker);

  ASSERT_MPR_2EXP (r, m);
}

static void
mpr_2exp__set_si (mpr_ptr w, long x, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_limb_t  xlimb;

  wp = RPTR(w);
  xlimb = (mp_limb_signed_t) x;
  wp[0] = xlimb & GMP_NUMB_MASK;
  msize = MRSIZE(m);

#if BITS_PER_ULONG <= GMP_NUMB_BITS
  MPN_FILL (wp+1, msize-1, (x >= 0 ? CNST_LIMB(0) : GMP_NUMB_MAX));

#else
  if (LIKELY (msize >= 2))
    {
      wp[1] = (xlimb >> GMP_NUMB_BITS)
        | (x >= 0 ? 0 : (GMP_NUMB_MAX << GMP_NAIL_BITS) & GMP_NUMB_MASK);
      MPN_FILL (wp+2, msize-2, (x >= 0 ? CNST_LIMB(0) : GMP_NUMB_MAX));
    }
#endif
}

static void
mpr_2exp__set_ui (mpr_ptr w, unsigned long x, mpm_srcptr m)
{
  mp_ptr     wp = RPTR(w);
  mp_size_t  msize = MRSIZE(m);

#if GMP_NUMB_BITS >= BITS_PER_ULONG
  wp[0] = x;
  MPN_ZERO_EXTEND (wp, 1, msize);

#else
  wp[0] = x & GMP_NUMB_MASK;
  wp[1] = x >> GMP_NUMB_BITS;

  if (MRSIZE(m) > 2)
    MPN_ZERO_EXTEND (wp, 2, msize);
#endif
}

static void
mpr_2exp__set_z (mpr_ptr w, mpz_srcptr x, mpm_srcptr m)
{
  mp_size_t  msize, xsize, xsize_signed;
  mp_srcptr  xp;
  mp_ptr     wp;

  wp = RPTR(w);
  xp = PTR(x);
  msize = MRSIZE(m);

  xsize_signed = SIZ(x);
  xsize = ABS (xsize_signed);
  xsize = MIN (xsize, msize);

  if (xsize_signed >= 0)
    {
      MPN_COPY (wp, xp, xsize);
      MPN_ZERO_EXTEND (wp, xsize, msize);
    }
  else
    {
      mpn_com_n (wp, xp, xsize);
      MPN_INCR_U (wp, xsize, CNST_LIMB(1));
      MPN_FILL_EXTEND (wp, xsize, msize, GMP_NUMB_MAX);
    }

  ASSERT_MPR_2EXP (w, m);
}

static void
mpr_2exp__sub (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  ASSERT_MPR_2EXP (x, m);
  ASSERT_MPR_2EXP (y, m);
  mpn_sub_n (RPTR(w), RPTR(x), RPTR(y), MRSIZE(m));
  ASSERT_MPR_2EXP (w, m);
}

static void
mpr_2exp__sub_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_size_t  msize;
  mp_limb_t  c;

  ASSERT_MPR_2EXP (x, m);

  wp = RPTR(w);
  xp = RPTR(x);
  msize = MRSIZE(m);

#if BITS_PER_ULONG > GMP_NUMB_BITS
  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      if (msize == 1)
        wp[0] = (xp[0] - y) & GMP_NUMB_MASK;
      else
        MPN_SUB_2 (c, wp, xp, msize,
                   (mp_limb_t) y >> GMP_NUMB_BITS,
                   (mp_limb_t) y & GMP_NUMB_MASK);
    }
  else
#endif
    {
      MPN_SUB_1 (c, wp, xp, msize, (mp_limb_t) y);
    }

  ASSERT_MPR_2EXP (w, m);
}

static void
mpr_2exp__submul_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_size_t  msize;

  ASSERT_MPR_2EXP (w, m);
  ASSERT_MPR_2EXP (x, m);
  ASSERT (MRALLOC(m) >= MRSIZE(m)+1);

  wp = RPTR(w);
  xp = RPTR(x);
  msize = MRSIZE(m);

#if BITS_PER_ULONG > GMP_NUMB_BITS
  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mp_limb_t  yp[2], c;
      ASSERT_CODE (wp[msize] = 0); /* so nail==0 */
      yp[0] = (mp_limb_t) y & GMP_NUMB_MASK;
      yp[1] = (mp_limb_t) y >> GMP_NUMB_BITS;
      MPN_SUBMUL_2 (c, wp, xp, msize, yp);
    }
  else
#endif
    {
      mpn_submul_1 (wp, xp, msize, (mp_limb_t) y);
    }

  ASSERT_MPR_2EXP (w, m);
}

static void
mpr_2exp__urandomm (mpr_ptr w, gmp_randstate_ptr rstate, mpm_srcptr m)
{
  _gmp_rand (RPTR(w), rstate, MBITS(m));
  ASSERT_MPR_2EXP (w, m);
}


void
mpm_init_2exp (mpm_ptr m, unsigned long n)
{
  mp_size_t      msize;
  unsigned long  highzeros;

  if (n < 1)
    DIVIDE_BY_ZERO;

  memset (m, '\0', sizeof (mpm_t));

  MBITS(m) = n;

  msize = BITS_TO_LIMBS (n);
  MRSIZE(m) = msize;
  MRALLOC(m) = 2*msize;

  mpz_init_set_ui (MZ(m), 0L);
  mpz_setbit (MZ(m), n);

  highzeros = msize * GMP_NUMB_BITS - n;
  MMASK(m) = GMP_NUMB_MASK >> highzeros;

  m->_mpr_add          = mpr_2exp__add;
  m->_mpr_add_ui       = mpr_2exp__add_ui;
  m->_mpr_addmul_ui    = mpr_2exp__addmul_ui;
  m->_mpr_clear        = mpr_gen__clear;
  m->_mpr_div          = mpr_2exp__div;
  m->_mpr_div_2exp     = mpr_2exp__div_2exp;
  m->_mpr_equal        = mpr_2exp__equal;
  m->_mpr_equal_one    = mpr_2exp__equal_one;
  m->_mpr_equal_negone = mpr_2exp__equal_negone;
  m->_mpr_equal_ui     = (msize == 1 ? mpr_2exp__equal_ui__1
                          :            mpr_2exp__equal_ui__2up);
  m->_mpr_equal_zero   = mpr_2exp__equal_zero;
  m->_mpr_gcd          = mpr_2exp__gcd;
  m->_mpr_get_ui       = mpr_2exp__get_ui;
  m->_mpr_get_z        = mpr_2exp__get_z;
  m->_mpr_invert       = mpr_2exp__invert;
  m->_mpr_jacobi       = mpr_2exp__jacobi;
  m->_mpr_mul          = mpr_2exp__mul;
  m->_mpr_mul_2exp     = mpr_2exp__mul_2exp;
  m->_mpr_mul_si       = mpr_2exp__mul_si;
  m->_mpr_mul_ui       = mpr_2exp__mul_ui;
  m->_mpr_neg          = mpr_2exp__neg;
  m->_mpr_new          = mpr_gen__new;
  m->_mpr_pow          = mpr_2exp__pow;
  m->_mpr_set          = mpr_gen__set;
  m->_mpr_set_si       = mpr_2exp__set_si;
  m->_mpr_set_ui       = mpr_2exp__set_ui;
  m->_mpr_set_z        = mpr_2exp__set_z;
  m->_mpr_sub          = mpr_2exp__sub;
  m->_mpr_sub_ui       = mpr_2exp__sub_ui;
  m->_mpr_submul_ui    = mpr_2exp__submul_ui;
  m->_mpr_swap         = mpr_gen__swap;
  m->_mpr_urandomm     = mpr_2exp__urandomm;
}
