/* Single-word REDC calculations.

   Copyright 2002, 2003, 2007 Kevin Ryde

   This file is part of Redc1.

   Redc1 is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.

   Redc1 is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with Redc1; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */




/* This is a spot of code for doing modular arithmetic within a single
   machine word (32 or 64 bits) using Peter Montgomery's REDC method
   ("Modular Multiplication Without Trial Division", Mathematics of
   Computation, volume 44, number 170, April 1985).

   The latest version of this code can be found at
   http://www.geocities.com/user42_kevin/redc1/index.html


   Building:

   There's nothing to build, the code is just the macros in this redc1.h
   file.  GNU MP longlong.h and gmp-impl.h are used for 2x1 word arithmetic
   so you need a GNU MP build directory, or at least the headers from it.
   Some of the resulting code may call helper functions in libgmp.  If
   that's the case then you'll need to link to the applicable libgmp version
   too.


   General:

   REDC works with an odd modulus N.  Inputs are converted to special redc
   residue form, and in that form the reduction part of a modular
   multiplication is done with two multiplies instead of a division.  This
   is somewhere between 2x and 10x faster, depending on the CPU, and that
   saving soon overcomes the cost of converting values into and out of redc
   form.

   The code style here is macros which take an output variable to store to
   and input variables expressions as operands, plus the modulus value and
   certain supporting pre-calculated values.  This style works with any
   compiler, though almost all the optimizations in longlong.h are GNU C
   specific, and thus GNU C is recommended.


   Usage:

   Start with a modulus N in a mp_limb_t, it must be an odd number

       mp_limb_t N = 12345;

   then use the various INIT macros to setup values dependent on that
   modulus.  You only need to make the ones needed for the macros you use.
   The full set is

       mp_limb_t  Ninv, RmodN, R2modN;
       REDC1_INIT_Ninv (Ninv, N);
       REDC1_INIT_RmodN (RmodN, N);
       REDC1_INIT_R2modN (R2modN, N, RmodN);

   Then convert input values to residue form using either REDC1_IN or
   REDC1_IN_SINGLE.  Inputs can be any limb value, they don't have to be in
   the range 0 to N-1.  REDC1_IN_SINGLE is designed for converting just one
   or two values, any more than that and REDC1_IN is faster.

       mp_limb_t x = 55555;
       mp_limb_t xR;
       REDC1_IN (xR, x, N, R2modN);
   or
       REDC1_IN_SINGLE (xR, x, N);

   Residues for the values 1 and -1 (ie. N-1) can be obtained from
   REDC1_IN_1 and REDC1_IN_NEG1 respectively.  1 is actually just the RmodN
   value initialized above, but using the REDC1_IN_1 macro can make it clear
   what's intended.  The residue of 0 is 0, there's no macro for that.

       mp_limb_t oneR, negoneR;
       REDC1_IN_1 (oneR, RmodN);
       REDC1_IN_NEG1 (negoneR, N, RmodN);

   Then REDC1_ADD, REDC1_SUB and REDC1_MUL can then be used to do
   arithmetic.  In each case the inputs are in residue form and the output
   likewise.

        mp_limb_t  zR;
        REDC1_NEG (zR, xR, N);
        REDC1_ADD (zR, xR, yR, N);
        REDC1_SUB (zR, xR, yR, N);
        REDC1_MUL (zR, xR, yR, N, Ninv);

   Results are converted back out of redc residue form with REDC1_OUT.  The
   output is always in the range 0 to N-1.

        mp_limb_t  z;
        REDC1_OUT (z, zR, N, Ninv);

   Residues themselves are always in the range 0 to N-1 too, which means the
   residue representation is unique and equality can be tested just with the
   usual == operator.


   Options:

   If you know you've got one (or more) spare high bits in the modulus N,
   ie. it fits within GMP_LIMB_BITS-1 (31 bits on a 32-bit system, or 63
   bits on a 64-bit system) then you can define REDC1_MODULUS_ONE_BIT_SPARE
   to select slightly faster versions of the macros.

        #define REDC1_MODULUS_ONE_BIT_SPARE  1
        #include "redc1.h"
        ...

   The gmp-impl.h ASSERT() macro is used.  You can define WANT_ASSERT to
   enable that code.  It causes significant slowdown and bloat, but is good
   when debugging.

        #define WANT_ASSERT  1
        #include "redc1.h"
        ...


   Notes:

   The R in the code is per Montgomery's paper, a power of 2 bigger than N,
   in this case always the machine word size, either 2^32 or 2^64.  The redc
   residue form is obtained by multiplying by a value R%N, that value is
   kept in the RmodN variable.

   Using explicit separate variables for each of the pre-calculated values
   Ninv, RmodN and R2modN is a little tedious, but gives maximum
   flexibility.  For example you can have two sets for two moduli at the
   same time, and let the compiler figure out when they go dead.  Or on a
   vector system you can have them in arrays, and vectorize loops doing the
   same operations on a set of moduli simultaneously.

   The pre-inversion method used by udiv_qrnnd_preinv() of gmp-impl.h is an
   alternative to redc.  It also does two multiplies instead of a division,
   like REDC, but the preinv has more supporting adds and subtracts, and can
   end up using more registers than confortably available on a starved CPU
   like i386.  Also for REDC the first of the two multiplies is just a
   low-half, which on some CPUs is a couple of cycles faster than getting a
   two-word result.


   Possible future enhancements:

   Make a REDC1_IN_SINGLE_PREINV using udiv_qrnnd_preinv.  */


#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"


#ifndef __REDC1_H__
#define __REDC1_H__


#if 0 /* change to 1 to print some traces */
#define REDC1_TRACE(expr) expr
#else
#define REDC1_TRACE(expr)
#endif


/*---------------------------------------------------------------------------*/
/* submod_dmsm()
   r = a-b mod m
   b=m is allowed, to support addmod_saam below */

#if REDC1_MODULUS_ONE_BIT_SPARE
/* When there's one bit spare we can detect underflow by looking at the sign
   bit of the result.  gcc knows how to turn "r < 0 ? m : 0" into an
   arithmetic shift to extract that bit, when the cpu has such an
   instruction.  */
#define submod_dmsm(r, a, b, m)                                               \
  do {                                                                        \
    mp_limb_t  __submod1__r;                                                  \
    ASSERT ((a) < (m));                                                       \
    ASSERT ((b) <= (m));                                                      \
    ASSERT ((m) < GMP_LIMB_HIGHBIT);                                          \
    __submod1__r = (a) - (b);                                                 \
    (r) = __submod1__r + ((mp_limb_signed_t) __submod1__r < 0 ? (m) : 0);     \
    ASSERT ((r) < (m));                                                       \
  } while (0)
#endif

#if ! defined (submod_dmsm)                             \
  && defined (__GNUC__) && ! defined (NO_ASM)           \
  && HAVE_HOST_CPU_FAMILY_x86 && GMP_LIMB_BITS == 32    \
  && HAVE_x86_cmov
/* Branch-free using cmov.  The generic C code below comes out using cmov,
   but only with an extra cmp instruction, gcc doesn't recognise that the
   carry flag from the subtract is already what's wanted.  */
#define submod_dmsm(r,a,b,m)                    \
  do {                                          \
    mp_limb_t  __submod1__r;                    \
    mp_limb_t  __submod1__add;                  \
    ASSERT ((a) < (m));                         \
    ASSERT ((b) <= (m));                        \
                                                \
    __submod1__add = 0;                         \
    __asm__ ("  subl   %2, %0\n"                \
             "  cmovc  %4, %1\n"                \
             : "=&r" (__submod1__r),            \
               "=r"  (__submod1__add)           \
             : "0"   (a),                       \
               "g"   (b),                       \
               "1"   (__submod1__add),          \
               "g"   (m));                      \
    (r) = __submod1__r + __submod1__add;        \
                                                \
    ASSERT ((r) < (m));                         \
  } while (0)
#endif

#if ! defined (submod_dmsm)                             \
  && defined (__GNUC__) && ! defined (NO_ASM)           \
  && HAVE_HOST_CPU_FAMILY_x86 && GMP_LIMB_BITS == 32
#define submod_dmsm(r,a,b,m)                    \
  do {                                          \
    ASSERT ((a) < (m));                         \
    ASSERT ((b) <= (m));                        \
    __asm__ ("  subl   %2, %0\n"                \
             "  jnc  " ASM_L(noadd) "\n"        \
             "  addl   %3, %0\n"                \
             ASM_L(noadd) ":\n"                 \
             : "=&r" (r)                        \
             : "0"   (a),                       \
               "g"   (b),                       \
               "g"   (m));                      \
                                                \
    ASSERT ((r) < (m));                         \
  } while (0)
#endif

#ifndef submod_dmsm
#define submod_dmsm(r, a, b, m)                                         \
  do {                                                                  \
    mp_limb_t  __submod1__a = (a);                                      \
    mp_limb_t  __submod1__r;                                            \
    ASSERT ((a) < (m));                                                 \
    ASSERT ((b) <= (m));                                                \
                                                                        \
    __submod1__r = __submod1__a - (b);                                  \
    (r) = __submod1__r + (__submod1__r > __submod1__a ? (m) : 0);       \
                                                                        \
    ASSERT ((r) < (m));                                                 \
  } while (0)
#endif


/*---------------------------------------------------------------------------*/
/* addmod_saam()
   r = a-b mod m */

#if REDC1_MODULUS_ONE_BIT_SPARE
/* With one bit spare there's no overflow from r=a+b, just a compare r>=m to
   reduce to 0<=r<m.  The r=a-(m-b) strategy for the general case below
   might be better on chips with good arithmetic shift but no conditional
   move.  */
#define addmod_saam(r, a, b, m)                                 \
  do {                                                          \
    mp_limb_t  __addmod1__r;                                    \
    ASSERT ((a) < (m));                                         \
    ASSERT ((b) < (m));                                         \
    ASSERT ((m) < GMP_LIMB_HIGHBIT);                            \
    __addmod1__r = (a) + (b);                                   \
    (r) = __addmod1__r - (__addmod1__r >= (m) ? (m) : 0);       \
    ASSERT ((r) < (m));                                         \
  } while (0)
#endif

/* The use of submod r=a-(m-b) here means just one compare test.  A plain
   r=a+b would need two: r<a for overflow and r>=m for reduction. */
#ifndef addmod_saam
#define addmod_saam(r, a, b, m)                         \
  do {                                                  \
    mp_limb_t  __addmod1__m = (m);                      \
    ASSERT ((a) < (m));                                 \
    ASSERT ((b) < (m));                                 \
    submod_dmsm (r, a, __addmod1__m - b, __addmod1__m);   \
  } while (0)
#endif


/*---------------------------------------------------------------------------*/
/* negmod_nam()
   r = -a mod m */

/* An alternative when REDC1_MODULUS_ONE_BIT_SPARE would be to look at the
   sign bit of -a, instead of testing a!=0.  Could help chips without a
   conditional move instruction.  */
#define negmod_nam(r, a, m)                                             \
  do {                                                                  \
    mp_limb_t  __negmod1__a = (a);                                      \
    ASSERT ((a) < (m));                                                 \
    (r) = (LIKELY (__negmod1__a != 0) ? (m) : 0) - __negmod1__a;        \
  } while (0)

/* r = -a mod m, for x!=0 */
#define negmod_nam_nonzero (r, a, m)    \
  do {                                  \
    ASSERT ((a) != 0);                  \
    ASSERT ((a) < (N));                 \
    (r) = (m) - (a);                    \
  } while (0)


/*---------------------------------------------------------------------------*/
/* umulmod_pmmd()
   r = x*y mod m */

#define umulmod_pmmd(r, x, y, m)                                        \
  do {                                                                  \
    mp_limb_t  __umulmod1__hi, __umulmod1__lo, __umulmod1__dummy_q;     \
    umul_ppmm (__umulmod1__hi, __umulmod1__lo, x, y);                   \
    udiv_qrnnd_unnorm (__umulmod1__dummy_q, r,                          \
                       __umulmod1__hi, __umulmod1__lo, m);              \
                                                                        \
    if (sizeof (unsigned long long) >= 2 * sizeof (mp_limb_t))          \
      ASSERT ((r) == ((unsigned long long) (x) * (y)) % (m));           \
  } while (0)

/* udiv_qrnnd_unnorm() is the same as udiv_qrnnd(), except it always
   allows unnormalized divisors */
#if UDIV_NEEDS_NORMALIZATION
#define udiv_qrnnd_unnorm(q, r, nh, nl, d)                      \
  do {                                                          \
    mp_limb_t  __udivun1__nh, __udivun1__nl;                    \
    mp_limb_t  __udivun1__d, __udivun1__r;                      \
    unsigned   __udivun1__shift;                                \
    ASSERT ((d) != 0);                                          \
    ASSERT ((nh) < (d));                                        \
    __udivun1__d = (d);                                         \
    count_leading_zeros (__udivun1__shift, __udivun1__d);       \
    __udivun1__d <<= __udivun1__shift;                          \
    __udivun1__nh = (nh);                                       \
    __udivun1__nl = (nl);                                       \
    __udivun1__nh <<= __udivun1__shift;                         \
    if (__udivun1__shift != 0)                                  \
      __udivun1__nh |=  (__udivun1__nl >>                       \
                         (GMP_LIMB_BITS - __udivun1__shift));   \
    __udivun1__nl <<= __udivun1__shift;                         \
    udiv_qrnnd (q, __udivun1__r,                                \
                __udivun1__nh, __udivun1__nl, __udivun1__d);    \
    (r) = __udivun1__r >> __udivun1__shift;                     \
  } while (0)
#else
#define udiv_qrnnd_unnorm(q, r, nh, nl, d)   udiv_qrnnd (q, r, nh, nl, d)
#endif


/*---------------------------------------------------------------------------*/
/* REDC1 initializers */

/* Calculate Ninv satisfying N*Ninv == 1 mod R */
#define REDC1_INIT_Ninv(Ninv, N)                \
  do {                                          \
    modlimb_invert (Ninv, N);                   \
    REDC1_TRACE (printf ("   Ninv=%#lX\n", Ninv));    \
  } while (0)

/* R mod N, calculated as (R-N) % N
   ENHANCE-ME: Maybe on some CPUs udiv_qrnnd is better than "%". */
#define REDC1_INIT_RmodN(RmodN, N)              \
  do {                                          \
    (RmodN) = (- (mp_limb_t) N) % N;            \
    REDC1_TRACE (printf ("   RmodN=%#lX\n", RmodN));  \
  } while (0)

/* R^2 mod N */
#define REDC1_INIT_R2modN(R2modN, N, RmodN)             \
  do {                                                  \
    umulmod_pmmd (R2modN, RmodN, RmodN, N);             \
    REDC1_TRACE (printf ("   R^2modN=%#lX\n", R2modN));       \
  } while (0)


/*---------------------------------------------------------------------------*/
/* REDC1 input */

/* t = x converted to redc form, using R2modN. */
#define REDC1_IN(xR, x, N, Ninv, R2modN)        \
  do {                                          \
    REDC1_MUL (xR, x, R2modN, N, Ninv);         \
  } while (0)

/* REDC1_IN_SINGLE sets xR to x converted into redc form, using only N.
   This is simply xR = x * R mod N.

   REDC1_IN needs an R2modN setup which costs mul+div, but then each input
   costs 2*mul.  REDC1_IN_SINGLE on the other hand is a div, or 2 divs if
   x>=N, each time.  So generally, as the name suggests, use REDC1_IN_SINGLE
   only if there's a single input to convert.  (Which is not as unlikely as
   it might sound, testing for pseudoprime to a given base, mod N, would
   convert just that base.)

   On a cpu with UDIV_NEEDS_NORMALIZATION, udiv_qrnnd_unnorm will end up
   doing a count_leading_zeros on N each time, but no great need to try to
   save that in a variable, because REDC1_IN_SINGLE is only meant to be for
   when doing just one or two input conversions anyway.  */

#define REDC1_IN_SINGLE(xR, x, N)                                       \
  do {                                                                  \
    mp_limb_t  __redc1in__dummy_q;                                      \
    mp_limb_t  __redc1in__x = (x);                                      \
    mp_limb_t  __redc1in__N = (N);                                      \
    if (UNLIKELY (__redc1in__x >= __redc1in__N))                        \
      __redc1in__x %= __redc1in__N;                                     \
    udiv_qrnnd_unnorm (__redc1in__dummy_q, xR,                          \
                       __redc1in__x, CNST_LIMB(0), __redc1in__N);       \
  } while (0)

/* set t = 1 in redc form, which is simply RmodN */
#define REDC1_IN_1(oneR, RmodN) \
  do {                          \
    (oneR) = (RmodN);           \
  } while (0)

/* set t = N-1 in redc form, which is simply N-RmodN */
#define REDC1_IN_NEG1(negoneR, N, RmodN)        \
  do {                                          \
    ASSERT (RmodN != 0);                        \
    (negoneR) = (N) - (RmodN);                  \
  } while (0)


/*---------------------------------------------------------------------------*/
/* REDC1 calculations */

/* REDC1_REDUCE does:
   m = (Th,Tl)*Ninv mod R, which means just m = Tl*Ninv mod R.
   t = (Th,Tl - m*N) / R.
   if (t < 0) return t+N else return t.
   Tl == low(q*N), so only need to calculate high of t, ie. Th-high(q*N). */
#define REDC1_REDUCE(t, Th, Tl, N, Ninv)                                \
  do {                                                                  \
    mp_limb_t  __redc1__hi, __redc1__lo;                                \
    ASSERT ((N) * (Ninv) == 1);                                         \
    /* must have  (Th,Tl)<R*N, so -N < t < N and only one addback */    \
    ASSERT ((Th) < (N));                                                \
    umul_ppmm (__redc1__hi,__redc1__lo, (Tl) * (Ninv), N);              \
    ASSERT (__redc1__lo == (Tl));                                       \
    submod_dmsm (t, Th, __redc1__hi, N);                                \
    ASSERT ((t) < (N));                                                 \
  } while (0)

/* z = x*y, in redc form.
   No assert on the size of x and y here, to allow for y==R2modN and x a
   full limb value from REDC1_IN.  The range check in REDC1_REDUCE is
   enough.  */
#define REDC1_MUL(zR, xR, yR, N, Ninv)                          \
  do {                                                          \
    mp_limb_t  __redc1mul__Th, __redc1mul__Tl;                  \
    umul_ppmm (__redc1mul__Th, __redc1mul__Tl, xR, yR);         \
    REDC1_REDUCE (zR, __redc1mul__Th, __redc1mul__Tl, N, Ninv); \
  } while (0)

/* z = x+y, all in redc form */
#define REDC1_ADD(zR, xR, yR, N)   \
  addmod_saam (zR, xR, yR, N)

/* z = x-y, all in redc form */
#define REDC1_SUB(zR, xR, yR, N)   \
  submod_dmsm (zR, xR, yR, N)

/* z = x-y, all in redc form */
#define REDC1_NEG(zR, xR, yR, N)   \
  negmod_nam (zR, xR, yR, N)

/* z = -x, for x!=0 */
#define REDC1_NEG_NONZERO(zR, xR, N)      \
  negmod_nam_nonzero (zR, xR, N)


/*---------------------------------------------------------------------------*/
/* REDC1 output */

/* Take a number out of redc form.  Same as REDC1_REDUCE(x, 0,xR, N,Ninv),
   but can use negmod_nam knowing Th==0. */
#define REDC1_OUT(x, xR, N, Ninv)                                       \
  do {                                                                  \
    mp_limb_t  __redc1out__hi, __redc1out__lo;                          \
    ASSERT ((xR) < (N));                                                \
    ASSERT ((N) * (Ninv) == 1);                                         \
    umul_ppmm (__redc1out__hi, __redc1out__lo, (xR) * (Ninv), N);       \
    ASSERT (__redc1out__lo == (xR));                                    \
    negmod_nam (x, __redc1out__hi, N);                                  \
  } while (0)



#endif /* __REDC1_H__ */
